// #define VERBOSE

/**
 * @file toy_multi_unstruct_2.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "yac_config.h"

#if defined YAC_NETCDF_ENABLED
#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include "utils.h"
#include "fields.h"
#include "yac_interface.h"
#include "geometry.h"
#include "vtk_output.h"
#include "read_mpiom_grid.h"
#include "test_function.h"

/* ------------------------------------------------- */

/* For simplicity we define the same 8 fields that are in the configuration file */

const char * fieldName[] = {"reg2d_conserv_field_out",
                            "unstruct_conserv_field_out",
                            "unstruct_2_conserv_field_out",
                            "cube_conserv_field_out",
                            "reg2d_avg_field_out",
                            "unstruct_avg_field_out",
                            "unstruct_2_avg_field_out",
                            "cube_avg_field_out",
                            "reg2d_hcsbb_field_out",
                            "unstruct_hcsbb_field_out",
                            "unstruct_2_hcsbb_field_out",
                            "cube_hcsbb_field_out",
                            "reg2d_2nd_conserv_field_out",
                            "unstruct_2nd_conserv_field_out",
                            "unstruct_2_2nd_conserv_field_out",
                            "cube_2nd_conserv_field_out",
                            "reg2d_rbf_field_out",
                            "unstruct_rbf_field_out",
                            "unstruct_2_rbf_field_out",
                            "cube_rbf_field_out"};

// redefine YAC assert macros
#undef YAC_ASSERT
#define STR_USAGE "Usage: %s -c configFilename -g gridFilename\n"
#define YAC_ASSERT(exp, msg) \
  { \
    if(!((exp))) { \
      fprintf(stderr, "ERROR: %s\n" STR_USAGE, msg, argv[0]); \
      exit(EXIT_FAILURE); \
    } \
  }

static void parse_arguments(
  int argc, char ** argv,
  char const ** configFilename, char const ** gridFilename);

int main (int argc, char *argv[]) {

  // Initialisation of MPI

  MPI_Init (0, NULL);

  /* The initialisation phase includes the reading of the configuration file */
#ifdef VERBOSE
  printf (". main: calling yac_cinit\n");
#endif

  double tic, toc, time;

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  char const * configFilename = "toy_multi.yaml"; // default configuration file
  char const * gridFilename = "TP04_lsm.nc"; // default grid file
  //char const * gridFilename = "TP10_lsm.nc";
  //char const * gridFilename = "GR30_lsm.nc";
  parse_arguments(argc, argv, &configFilename, &gridFilename);
  yac_cinit ();
  yac_cread_config_yaml(configFilename);

  /* The usual component definition, here for two sequential components on the same process */

#ifdef VERBOSE
  printf (". main: calling yac_cdef_comp\n");
#endif

  int comp_id;
  char * comp_name = "MPIOM";

  yac_cdef_comp ( comp_name, &comp_id );
#ifdef VERBOSE
  printf ( ". main: defined %s with local comp ID %i \n", "MPIOM", comp_id );
#endif

  MPI_Comm local_comm;

  yac_cget_comp_comm(comp_id, &local_comm);

  int rank, size;
  MPI_Comm_rank(local_comm,&rank);
  MPI_Comm_size(local_comm,&size);
  MPI_Comm_free(&local_comm);


  int nbr_fields;
  int cell_point_id;
  int corner_point_id;

  int field_ids[20];
  int grid_id;

  /* Grid definition for the third component*/

  int nbr_vertices;
  int nbr_cells;
  int * num_vertices_per_cell;
  int * cell_to_vertex;

  int * cell_mask;

  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * cell_core_mask;
  int * global_cell_id;
  int * corner_core_mask;
  int * global_corner_id;

  read_part_mpiom_grid_information(gridFilename, &nbr_vertices, &nbr_cells,
                                   &num_vertices_per_cell, &cell_to_vertex,
                                   &x_vertices, &y_vertices, &x_cells,
                                   &y_cells, &global_cell_id,
                                   &cell_mask,
                                   &cell_core_mask, &global_corner_id,
                                   &corner_core_mask, rank, size);

  double * x_points, * y_points;

  x_points = x_vertices;
  y_points = y_vertices;

  double * x_center, * y_center;

  x_center = x_cells;
  y_center = y_cells;

  yac_cdef_grid_unstruct ( "mpiom_grid", nbr_vertices, nbr_cells, num_vertices_per_cell,
                           x_vertices, y_vertices, cell_to_vertex, &grid_id );

  yac_cset_global_index(global_cell_id, CELL, grid_id);
  yac_cset_core_mask(cell_core_mask, CELL, grid_id);
  yac_cset_global_index(global_corner_id, CORNER, grid_id);
  yac_cset_core_mask(corner_core_mask, CORNER, grid_id);

  yac_cdef_points_unstruct ( grid_id, nbr_vertices, CORNER, x_points, y_points, &corner_point_id );
  yac_cdef_points_unstruct ( grid_id, nbr_cells, CELL, x_center, y_center, &cell_point_id );

  yac_cset_mask ( cell_mask, cell_point_id );

  /* Field definition for the third component */

  nbr_fields = 0;
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(
      fieldName[i+0], comp_id, &cell_point_id, 1, 1, "2", YAC_TIME_UNIT_SECOND,
      &field_ids[nbr_fields++]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(
      fieldName[i+4], comp_id, &corner_point_id, 1, 1, "2", YAC_TIME_UNIT_SECOND,
      &field_ids[nbr_fields++]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(
      fieldName[i+8], comp_id, &cell_point_id, 1, 1, "2", YAC_TIME_UNIT_SECOND,
      &field_ids[nbr_fields++]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(
      fieldName[i+12], comp_id, &cell_point_id, 1, 1, "2", YAC_TIME_UNIT_SECOND,
      &field_ids[nbr_fields++]);
  for (unsigned i = 0; i < 4; ++i)
    yac_cdef_field(
      fieldName[i+16], comp_id, &cell_point_id, 1, 1, "2", YAC_TIME_UNIT_SECOND,
      &field_ids[nbr_fields++]);

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("toy_multi_unstruct_2: Time for initialisation %f\n", time );

  /* Search. */

#ifdef VERBOSE
  printf (". main: calling yac_cenddef\n");
#endif

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  yac_cenddef ( );

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("toy_multi_unstruct_2: Time for search %f\n", time );

  double * conserv_in_reg2d_point_data = xmalloc(nbr_cells * sizeof(*conserv_in_reg2d_point_data));
  double * conserv_in_unstruct_point_data = xmalloc(nbr_cells * sizeof(*conserv_in_unstruct_point_data));
  double * conserv_in_cube_point_data = xmalloc(nbr_cells * sizeof(*conserv_in_cube_point_data));

  double * avg_in_reg2d_point_data = xmalloc(nbr_vertices * sizeof(*avg_in_reg2d_point_data));
  double * avg_in_unstruct_point_data = xmalloc(nbr_vertices * sizeof(*avg_in_unstruct_point_data));
  double * avg_in_cube_point_data = xmalloc(nbr_vertices * sizeof(*avg_in_cube_point_data));

  double * hcsbb_in_reg2d_point_data = xmalloc(nbr_cells * sizeof(*hcsbb_in_reg2d_point_data));
  double * hcsbb_in_unstruct_point_data = xmalloc(nbr_cells * sizeof(*hcsbb_in_unstruct_point_data));
  double * hcsbb_in_cube_point_data = xmalloc(nbr_cells * sizeof(*hcsbb_in_cube_point_data));

  double * conserv_2nd_in_reg2d_point_data = xmalloc(nbr_cells * sizeof(*conserv_2nd_in_reg2d_point_data));
  double * conserv_2nd_in_unstruct_point_data = xmalloc(nbr_cells * sizeof(*conserv_2nd_in_unstruct_point_data));
  double * conserv_2nd_in_cube_point_data = xmalloc(nbr_cells * sizeof(*conserv_2nd_in_cube_point_data));

  double * rbf_in_reg2d_point_data = xmalloc(nbr_cells * sizeof(*rbf_in_reg2d_point_data));
  double * rbf_in_unstruct_point_data = xmalloc(nbr_cells * sizeof(*rbf_in_unstruct_point_data));
  double * rbf_in_cube_point_data = xmalloc(nbr_cells * sizeof(*rbf_in_cube_point_data));

  for (int i = 0; i < nbr_cells; ++i) {
    conserv_in_reg2d_point_data[i] = -10;
    conserv_in_unstruct_point_data[i] = -10;
    conserv_in_cube_point_data[i] = -10;

    hcsbb_in_reg2d_point_data[i] = -10;
    hcsbb_in_unstruct_point_data[i] = -10;
    hcsbb_in_cube_point_data[i] = -10;

    conserv_2nd_in_reg2d_point_data[i] = -10;
    conserv_2nd_in_unstruct_point_data[i] = -10;
    conserv_2nd_in_cube_point_data[i] = -10;

    rbf_in_reg2d_point_data[i] = -10;
    rbf_in_unstruct_point_data[i] = -10;
    rbf_in_cube_point_data[i] = -10;
  }

  for (int i = 0; i < nbr_vertices; ++i) {
    avg_in_reg2d_point_data[i] = -10;
    avg_in_unstruct_point_data[i] = -10;
    avg_in_cube_point_data[i] = -10;
  }

  double * cell_out_point_data = xmalloc(nbr_cells * sizeof(*cell_out_point_data));
  double * corner_out_point_data = xmalloc(nbr_vertices * sizeof(*corner_out_point_data));

  int err;
  int info;

  int cell_to_vertex_offset = 0;

  for (int i = 0; i < nbr_cells; ++i) {

    double middle_point[3] = {0, 0, 0};

    for (int j = 0; j < num_vertices_per_cell[i]; ++j) {

      double curr_point[3];

      LLtoXYZ(x_vertices[cell_to_vertex[cell_to_vertex_offset + j]],
              y_vertices[cell_to_vertex[cell_to_vertex_offset + j]],
              curr_point);

      middle_point[0] += curr_point[0];
      middle_point[1] += curr_point[1];
      middle_point[2] += curr_point[2];
    }

    double scale = 1.0 / sqrt(middle_point[0] * middle_point[0] +
                              middle_point[1] * middle_point[1] +
                              middle_point[2] * middle_point[2]);

    middle_point[0] *= scale;
    middle_point[1] *= scale;
    middle_point[2] *= scale;

    double lon, lat;

    XYZtoLL(middle_point, &lon, &lat);

    cell_to_vertex_offset += num_vertices_per_cell[i];

    cell_out_point_data[i] = test_func(lon, lat);
  }

  for (int i = 0; i < nbr_vertices; ++i)
    corner_out_point_data[i] = test_func(x_points[i], y_points[i]);

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  {
    {
      double *point_set_data[1] = {cell_out_point_data};
      double **collection_data[1] = {point_set_data};

      yac_cput(field_ids[2], 1, collection_data, &info, &err);
      yac_cput(field_ids[10], 1, collection_data, &info, &err);
      yac_cput(field_ids[14], 1, collection_data, &info, &err);
      yac_cput(field_ids[18], 1, collection_data, &info, &err);
    }
    {
      double *point_set_data[1] = {corner_out_point_data};
      double **collection_data[1] = {point_set_data};

      point_set_data[0] = corner_out_point_data;
      yac_cput(field_ids[6], 1, collection_data, &info, &err);
    }
  }

  {
    double *collection_data[1];

    collection_data[0] = conserv_in_reg2d_point_data;
    yac_cget(field_ids[0], 1, collection_data, &info, &err);
    collection_data[0] = conserv_in_unstruct_point_data;
    yac_cget(field_ids[1], 1, collection_data, &info, &err);
    collection_data[0] = conserv_in_cube_point_data;
    yac_cget(field_ids[3], 1, collection_data, &info, &err);

    collection_data[0] = avg_in_reg2d_point_data;
    yac_cget(field_ids[4], 1, collection_data, &info, &err);
    collection_data[0] = avg_in_unstruct_point_data;
    yac_cget(field_ids[5], 1, collection_data, &info, &err);
    collection_data[0] = avg_in_cube_point_data;
    yac_cget(field_ids[7], 1, collection_data, &info, &err);

    collection_data[0] = hcsbb_in_reg2d_point_data;
    yac_cget(field_ids[8], 1, collection_data, &info, &err);
    collection_data[0] = hcsbb_in_unstruct_point_data;
    yac_cget(field_ids[9], 1, collection_data, &info, &err);
    collection_data[0] = hcsbb_in_cube_point_data;
    yac_cget(field_ids[11], 1, collection_data, &info, &err);

    collection_data[0] = conserv_2nd_in_reg2d_point_data;
    yac_cget(field_ids[12], 1, collection_data, &info, &err);
    collection_data[0] = conserv_2nd_in_unstruct_point_data;
    yac_cget(field_ids[13], 1, collection_data, &info, &err);
    collection_data[0] = conserv_2nd_in_cube_point_data;
    yac_cget(field_ids[15], 1, collection_data, &info, &err);

    collection_data[0] = rbf_in_reg2d_point_data;
    yac_cget(field_ids[16], 1, collection_data, &info, &err);
    collection_data[0] = rbf_in_unstruct_point_data;
    yac_cget(field_ids[17], 1, collection_data, &info, &err);
    collection_data[0] = rbf_in_cube_point_data;
    yac_cget(field_ids[19], 1, collection_data, &info, &err);
  }

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("toy_multi_unstruct_2: Time for ping-pong %f\n", time );

  //----------------------------------------------------------
  // write field to vtk output file
  //----------------------------------------------------------

  char vtk_filename[32];

  sprintf(vtk_filename, "unstruct_2_out_%d.vtk", rank);

  double point_data[nbr_vertices][3];
  for (int i = 0; i < nbr_vertices; ++i) {
   LLtoXYZ(x_vertices[i], y_vertices[i], point_data[i]);
  }

  VTK_FILE *vtk_file = vtk_open(vtk_filename, "unstruct_2_out");
  vtk_write_point_data(vtk_file, (double *)point_data, nbr_vertices);
  vtk_write_cell_data(vtk_file, (unsigned *)cell_to_vertex,
                      (unsigned*)num_vertices_per_cell, nbr_cells);
  vtk_write_point_scalars_int(vtk_file, corner_core_mask, nbr_vertices,
                              "corner_core_mask");
  vtk_write_point_scalars_int(vtk_file, global_corner_id, nbr_vertices,
                              "global_corner_id");
  vtk_write_cell_scalars_int(vtk_file, cell_core_mask, nbr_cells,
                             "cell_core_mask");
  vtk_write_cell_scalars_int(vtk_file, global_cell_id, nbr_cells,
                             "global_cell_id");

  vtk_write_cell_scalars_double(vtk_file, conserv_in_reg2d_point_data, nbr_cells,
                                "conserv_in_reg2d");
  vtk_write_cell_scalars_double(vtk_file, conserv_in_unstruct_point_data, nbr_cells,
                                "conserv_in_unstruct");
  vtk_write_cell_scalars_double(vtk_file, conserv_in_cube_point_data, nbr_cells,
                                "conserv_in_cube");

  vtk_write_point_scalars_double(vtk_file, avg_in_reg2d_point_data, nbr_vertices,
                                 "avg_in_reg2d");
  vtk_write_point_scalars_double(vtk_file, avg_in_unstruct_point_data, nbr_vertices,
                                 "avg_in_unstruct");
  vtk_write_point_scalars_double(vtk_file, avg_in_cube_point_data, nbr_vertices,
                                 "avg_in_cube");

  vtk_write_cell_scalars_double(vtk_file, hcsbb_in_reg2d_point_data, nbr_cells,
                                "hcsbb_in_reg2d");
  vtk_write_cell_scalars_double(vtk_file, hcsbb_in_unstruct_point_data, nbr_cells,
                                "hcsbb_in_unstruct");
  vtk_write_cell_scalars_double(vtk_file, hcsbb_in_cube_point_data, nbr_cells,
                                "hcsbb_in_cube");

  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_reg2d_point_data, nbr_cells,
                                "2nd_conserv_in_reg2d");
  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_unstruct_point_data, nbr_cells,
                                "2nd_conserv_in_unstruct");
  vtk_write_cell_scalars_double(vtk_file, conserv_2nd_in_cube_point_data, nbr_cells,
                                "2nd_conserv_in_cube");

  vtk_write_cell_scalars_double(vtk_file, rbf_in_reg2d_point_data, nbr_cells,
                                "rbf_in_reg2d");
  vtk_write_cell_scalars_double(vtk_file, rbf_in_unstruct_point_data, nbr_cells,
                                "rbf_in_unstruct");
  vtk_write_cell_scalars_double(vtk_file, rbf_in_cube_point_data, nbr_cells,
                                "rbf_in_cube");

  vtk_write_point_scalars_double(vtk_file, corner_out_point_data, nbr_vertices,
                                 "corner_out");
  vtk_write_cell_scalars_double(vtk_file, cell_out_point_data, nbr_cells, "cell_out");
  vtk_close(vtk_file);

  yac_cfinalize();

  MPI_Finalize();

  free(conserv_in_reg2d_point_data);
  free(conserv_in_unstruct_point_data);
  free(conserv_in_cube_point_data);
  free(avg_in_reg2d_point_data);
  free(avg_in_unstruct_point_data);
  free(avg_in_cube_point_data);
  free(hcsbb_in_reg2d_point_data);
  free(hcsbb_in_unstruct_point_data);
  free(hcsbb_in_cube_point_data);
  free(conserv_2nd_in_reg2d_point_data);
  free(conserv_2nd_in_unstruct_point_data);
  free(conserv_2nd_in_cube_point_data);
  free(rbf_in_reg2d_point_data);
  free(rbf_in_unstruct_point_data);
  free(rbf_in_cube_point_data);
  free(corner_out_point_data);
  free(cell_out_point_data);

  free(num_vertices_per_cell);
  free(cell_to_vertex);
  free(cell_mask);
  free(x_vertices);
  free(y_vertices);
  free(x_cells);
  free(y_cells);
  free(cell_core_mask);
  free(global_cell_id);
  free(corner_core_mask);
  free(global_corner_id);

  return EXIT_SUCCESS;
}

static void parse_arguments(
  int argc, char ** argv,
  char const ** configFilename, char const ** gridFilename) {

  int opt;
  while ((opt = getopt(argc, argv, "c:g:")) != -1) {
    YAC_ASSERT((opt == 'c') || (opt == 'g'), "invalid command argument")
    switch (opt) {
      default:
      case 'c':
        *configFilename = optarg;
        break;
      case 'g':
        *gridFilename = optarg;
        break;
    }
  }
}

#else
#include <stdlib.h>
#include <stdio.h>
int main () {
  printf ("Examples requires compiling with NetCDF.\n");
  return EXIT_FAILURE;
}
#endif
